import numpy as np


class Meter(object):
    '''Meters provide a way to keep track of important statistics in an online manner.
    This class is abstract, but provides a standard interface for all meters to follow.
    '''

    def reset(self):
        '''Resets the meter to default settings.'''
        pass

    def add(self, value):
        '''Log a new value to the meter
        Args:
            value: Next restult to include.
        '''
        pass

    def value(self):
        '''Get the value of the meter in the current state.'''
        pass


class AverageValueMeter(Meter):
    def __init__(self):
        super(AverageValueMeter, self).__init__()
        self.reset()
        self.val = 0

    def add(self, value, n=1):
        self.val = value
        self.sum += value
        self.var += value * value
        self.n += n

        if self.n == 0:
            self.mean, self.std = np.nan, np.nan
        elif self.n == 1:
            self.mean = 0.0 + self.sum  # This is to force a copy in torch/numpy
            self.std = np.inf
            self.mean_old = self.mean
            self.m_s = 0.0
        else:
            self.mean = self.mean_old + (value - n * self.mean_old) / float(self.n)
            self.m_s += (value - self.mean_old) * (value - self.mean)
            self.mean_old = self.mean
            self.std = np.sqrt(self.m_s / (self.n - 1.0))

    def value(self):
        return self.mean, self.std

    def reset(self):
        self.n = 0
        self.sum = 0.0
        self.var = 0.0
        self.val = 0.0
        self.mean = np.nan
        self.mean_old = 0.0
        self.m_s = 0.0
        self.std = np.nan



from . import base
from . import functional as F
from .base import Activation


class IoU(base.Metric):
    __name__ = 'iou_score'

    def __init__(self, eps=1e-7, threshold=0.5, activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps
        self.threshold = threshold
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        y_pr = self.activation(y_pr)
        return F.iou(
            y_pr, y_gt,
            eps=self.eps,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


class Fscore(base.Metric):

    def __init__(self, beta=1, eps=1e-7, threshold=0.5, activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps
        self.beta = beta
        self.threshold = threshold
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        y_pr = self.activation(y_pr)
        return F.f_score(
            y_pr, y_gt,
            eps=self.eps,
            beta=self.beta,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


class Accuracy(base.Metric):

    def __init__(self, threshold=0.5, activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.threshold = threshold
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        y_pr = self.activation(y_pr)
        return F.accuracy(
            y_pr, y_gt,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


class Recall(base.Metric):

    def __init__(self, eps=1e-7, threshold=0.5, activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps
        self.threshold = threshold
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        y_pr = self.activation(y_pr)
        return F.recall(
            y_pr, y_gt,
            eps=self.eps,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


class Precision(base.Metric):

    def __init__(self, eps=1e-7, threshold=0.5, activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps
        self.threshold = threshold
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        y_pr = self.activation(y_pr)
        return F.precision(
            y_pr, y_gt,
            eps=self.eps,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )



import sys, os
sys.path.insert(0, os.path.abspath('.'))
import numpy as np
import matplotlib.pyplot as plt
import mammopy as mi


class visualization():


    def visualize_breast_dense_segmentation(image, breast_prediction_mask, dense_prediction_mask):
        
        """
        Computes the plots of mammogram/s.
        Args:
            image: Input image in np.ndarray
            breast_prediction_mask: Customize output to visualize Breast Area
            dense_prediction_mask: Customize output to visualize Dense Area
        Returns:
            Three plots:
                - Combined plot of breast and dense area.
                - Plot of breast area.
                - Plot of dense area.
        Raises:
            TypeError: If `image` is not in nd.narray format.
            TypeError: If `breast_prediction_mask` is not in nd.narray format.
            TypeError: If `dense_prediction_mask` is not in nd.narray format.
        """
        
        
        if (type(image) not in [np.ndarray]) and (type(breast_prediction_mask) not in [np.ndarray]) and (type(dense_prediction_mask) not in [np.ndarray]):
            raise TypeError("All Inputs must be np.ndarray")
        edges = mi.analysis.canny_edges(breast_prediction_mask)
        #plotttig the results
        combined_sigm, axes = plt.subplots(1,2, figsize = (15,10),squeeze=False)
        axes[0, 0].set_title('Image', fontsize=16)
        axes[0, 1].set_title("Breast and dense tissue segmentation", fontsize=20)
        axes[0, 0].imshow(image, cmap='gray')
        axes[0, 0].set_axis_off()
        axes[0, 1].imshow(image, cmap='gray')
        axes[0, 1].imshow(mi.analysis.mask_to_rgba(edges, color='red'), cmap='gray')
        axes[0, 1].imshow(mi.analysis.mask_to_rgba(dense_prediction_mask, color='green'), cmap='gray', alpha=0.7)
        axes[0, 1].set_axis_off()

        brest_sigm, axes = plt.subplots(1,1, squeeze=False)
        axes[0, 0].imshow(image, cmap='gray')
        axes[0, 0].imshow(mi.analysis.mask_to_rgba(edges, color='red'), cmap='gray')
        axes[0, 0].set_axis_off()

        dens_sigm, axes = plt.subplots(1,1, squeeze=False)
        axes[0, 0].imshow(image, cmap='gray')
        axes[0, 0].imshow(mi.analysis.mask_to_rgba(dense_prediction_mask, color='green'), cmap='gray', alpha=0.7)
        axes[0, 0].set_axis_off()        

        return combined_sigm, brest_sigm, dens_sigm



import sys, os
sys.path.insert(0, os.path.abspath('.'))
import numpy as np
from skimage import feature
import torch
import matplotlib.pyplot as plt
import pydicom
from PIL import Image
import mammopy as mi


class analysis():
        
    MASK_COLORS = ["red", "green", "blue", "yellow", "magenta", "cyan"]
    def mask_to_rgba(mask, color="red"):
        """
        Converts binary segmentation mask from white to red color.
        Also adds alpha channel to make black background transparent.
        Args:
            mask (numpy.ndarray): [description]
            color (str, optional): Check `MASK_COLORS` for available colors. Defaults to "red".
        Returns:
            numpy.ndarray: [description]
        """
        assert color in analysis.MASK_COLORS
        assert mask.ndim == 3 or mask.ndim == 2

        h = mask.shape[0]
        w = mask.shape[1]
        zeros = np.zeros((h, w))
        ones = mask.reshape(h, w)
        if color == "red":
            return np.stack((ones, zeros, zeros, ones), axis=-1)
        elif color == "green":
            return np.stack((zeros, ones, zeros, ones), axis=-1)
        elif color == "blue":
            return np.stack((zeros, zeros, ones, ones), axis=-1)
        elif color == "yellow":
            return np.stack((ones, ones, zeros, ones), axis=-1)
        elif color == "magenta":
            return np.stack((ones, zeros, ones, ones), axis=-1)
        elif color == "cyan":
            return np.stack((zeros, ones, ones, ones), axis=-1)

    def density_estimation(model, img):
        """
        Computes the percentage of mammogram density in an input image.
        Args:
            model (str): The name of the pre-trained model to use.
            img: Input image in np.ndarray or PIL.Image format
            
        Returns:
            - image: The image in 256*256 size used for analysis
            - pred1: The breast area of mammogram
            - pred2: The dense area of mammogram
            - percentage_density: The percentage of mammogram density in the input image.
        Raises:
            TypeError: If `model` is not a string or `image_path` is not a string.
            ValueError: If `model` is not a valid pre-trained model name or `image_path`
            does not point to a valid image file.
        """
        
        if not isinstance(model, torch.nn.Module):
            raise TypeError("Model must be a PyTorch model.")
        if type(img) not in [np.ndarray, Image.Image]:
            raise TypeError("Input must be np.ndarray or PIL.Image")
        
        
        img = mi.pre_process.image_tensor(img)
        # Compute the mammogram density
        
        pred1, pred2 = model.module.predict(img)

        img = img[0].cpu().numpy().transpose(1, 2, 0)
        img = img[:, :, 0]

        pred1 = pred1[0].cpu().numpy().transpose(1, 2, 0)
        pred1 = pred1[:, :, 0]

        pred2 = pred2[0].cpu().numpy().transpose(1, 2, 0)
        pred2 = pred2[:, :, 0]

        breast_area = np.sum(np.array(pred1) == 1)
        dense_area = np.sum(np.array(pred2) == 1)
        density = (dense_area / breast_area) * 100
        density = np.rint(density)

        return img, pred1, pred2, density


    def single_density_estimation(model, img, visualization = False, breast_segmentation=False, dense_segmentation=False):
        """
        Computes the percentage of mammogram density in an input image.
        Args:
            model (str): The name of the pre-trained model to use.
            img: Input image in np.ndarray or PIL.Image format
            visualization: For display output
            breast_segmentation: Customize output to visualize Breast Area
            dense_segmentation: Customize output to visualize Dense Area
        Returns:
            dict: A dictionary containing the results of the computation. The dictionary
            has the following keys:
                - non_dense_area: The total number of pixels in the input image that corresponse 
                to non-dense tissue. (i.e., non-fibroglandular)
                - dense_area: The total number of pixels in the input image that corresponse
                to dense tissue. (i.e., fibroglandular)
                - percentage_density: The percentage of mammogram density in the input image.
        Raises:
            TypeError: If `model` is not a string or `image_path` is not a string.
            ValueError: If `model` is not a valid pre-trained model name or `image_path`
            does not point to a valid image file.
        """
        if not isinstance(model, torch.nn.Module):
            raise TypeError("Model must be a PyTorch model.")
        if type(img) not in [np.ndarray, Image.Image]:
            raise TypeError("Input must be np.ndarray or PIL.Image")
        elif (visualization != True) and ((breast_segmentation == True) or (dense_segmentation == True)):
            raise ValueError("For display option Visualization must be true")
        
        # Compute the mammogram density
        result = {}

        img, pred1, pred2, _ = mi.analysis.density_estimation(model, img)

        breast_area = np.sum(np.array(pred1) == 1)
        dense_area = np.sum(np.array(pred2) == 1)
        density = (dense_area / breast_area) * 100
        density = np.rint(density)
        
        # Populate the result dictionary with the computed values
        result["Non_Dense_Area"] = breast_area
        result["Dense_Area"] = dense_area
        result["Percentage_Density"] = density

        if visualization == False:
            return result
        elif (visualization == True) and (breast_segmentation == True) and (dense_segmentation == False):
            edges = analysis.canny_edges(pred1)
            plt.title('Breast area contour')
            plt.imshow(img, cmap='gray')
            plt.imshow(analysis.mask_to_rgba(edges, color='red'), cmap='gray')
            plt.axis('off')
            plt.show()
            return result
        elif (visualization == True) and (breast_segmentation == False) and (dense_segmentation == True):
            plt.title('Dense tissues')
            plt.imshow(img, cmap='gray')
            plt.imshow(analysis.mask_to_rgba(pred2, color='green'), cmap='gray')
            plt.axis('off')
            plt.show()
            return result
        else:
            edges = analysis.canny_edges(pred1)
            fig, axes = plt.subplots(1,2, figsize = (15,10),squeeze=False)
            axes[0, 0].set_title('Image', fontsize=16)
            axes[0, 1].set_title("Breast and dense tissue segmentation", fontsize=20)
            axes[0, 0].imshow(img, cmap='gray')
            axes[0, 0].set_axis_off()
            axes[0, 1].imshow(img, cmap='gray')
            axes[0, 1].imshow(analysis.mask_to_rgba(edges, color='red'), cmap='gray')
            axes[0, 1].imshow(analysis.mask_to_rgba(pred2, color='green'), cmap='gray', alpha=0.7)
            axes[0, 1].set_axis_off()
            plt.show()
            return result

    def canny_edges(image_array):
        """
        Detect edges using the Canny algorithm.
        
        Parameters:
        image_array (numpy.ndarray): The input image array
        
        Returns:
        edges (numpy.ndarray): A binary edge map with detected edges marked with 1.
        
        Raises:
        TypeError: If the input image is not a numpy array.
        """
        if not isinstance(image_array, np.ndarray):
            raise TypeError("Input image must be a numpy array.")
        # Detect edges using Canny algorithm with sigma value of 3.
        edges = feature.canny(image_array, sigma=3)
        return edges

    def multi_density_estimation(model, folder_path):

        """
        Computes the percentage of mammogram density in an input image.
        Args:
            model (str): The name of the pre-trained model to use.
            folder_path (str): The path to the input folder of files.
        Returns:
            dict: A dictionary containing the results of the computation of each valid file. The dictionary
            has the following keys:
                - File: Name of the valid file, Skips the analysis for wrong file format
                - Non_Dense_Area: The total number of pixels in the input file that corresponse 
                to non-dense tissue. (i.e., non-fibroglandular)
                - Dense_Area: The total number of pixels in the input file that corresponse
                to dense tissue. (i.e., fibroglandular)
                - Percentage_Density: The percentage of mammogram density of the input files.
        Raises:
            TypeError: If `model` is not a string or `folder_path` is not a string.
            ValueError: If `model` is not a valid pre-trained model name or `folder_path`
            does not point to a valid folder.
        """
        if not isinstance(model, torch.nn.Module):
            raise TypeError("Model must be a PyTorch model.")
    
        if os.path.exists(folder_path):
            files = os.scandir(folder_path)
        else:
            raise FileExistsError("Folder path must be a string with '/' forward slashs")
        result = []
        for file in files:
            if file.is_file():
                file_ext = file.name.split(".")[-1]
                if file_ext in ['jpg', 'JPG', 'jpeg', 'JPEG', 'png', 'PNG', 'tif', 'tiff', 'TIF', 'TIFF']:
                    print(file.name)
                    img = Image.open(open(file, 'rb')).convert('RGB')
                    temp = analysis.single_density_estimation(model=model, img=img)
                elif file_ext == 'dcm' or file_ext == 'DCM':
                    dicom_file = pydicom.dcmread(file)
                    pil_image = Image.fromarray(dicom_file.pixel_array)
                    img = pil_image.convert('RGB')
                    temp = analysis.single_density_estimation(model=model, img=img)
                else:
                    try:
                        ds = pydicom.dcmread(file)
                        pixel_array  = ds.pixel_array
                        if ds.PhotometricInterpretation == 'MONOCHROME1':
                            pixel_array = (pixel_array * 255.0 / np.max(pixel_array)).astype(np.uint8)
                            pil_image = Image.fromarray(pixel_array)
                            img = pil_image.convert('RGB')
                            img = np.max(img) - img
                            np_to_PILimg = Image.fromarray(img)
                            temp = analysis.single_density_estimation(model=model, img=np_to_PILimg)
                        else:
                            temp = None
                            pass
                    except pydicom.errors.InvalidDicomError:
                        temp = None
                        pass
        # Populate the result dictionary with the computed values
            if temp is None:
                pass
            else:
                temp = {'File': file.name, "Non_Dense_Area": temp["Non_Dense_Area"], 'Dense_Area': temp["Dense_Area"], 'Percentage_Density': temp["Percentage_Density"]}
                result.append(temp)
        return result




""" Each encoder should have following attributes and methods and be inherited from `_base.EncoderMixin`

Attributes:

    _out_channels (list of int): specify number of channels for each encoder feature tensor
    _depth (int): specify number of stages in decoder (in other words number of downsampling operations)
    _in_channels (int): default number of input channels in first Conv2d layer for encoder (usually 3)

Methods:

    forward(self, x: torch.Tensor)
        produce list of features of different spatial resolutions, each feature is a 4D torch.tensor of
        shape NCHW (features should be sorted in descending order according to spatial resolution, starting
        with resolution same as input `x` tensor).

        Input: `x` with shape (1, 3, 64, 64)
        Output: [f0, f1, f2, f3, f4, f5] - features with corresponding shapes
                [(1, 3, 64, 64), (1, 64, 32, 32), (1, 128, 16, 16), (1, 256, 8, 8),
                (1, 512, 4, 4), (1, 1024, 2, 2)] (C - dim may differ)

        also should support number of features according to specified depth, e.g. if depth = 5,
        number of feature tensors = 6 (one with same resolution as input and 5 downsampled),
        depth = 3 -> number of feature tensors = 4 (one with same resolution as input and 3 downsampled).
"""

import torchvision
import torch.nn as nn

from ._base import EncoderMixin


class MobileNetV2Encoder(torchvision.models.MobileNetV2, EncoderMixin):

    def __init__(self, out_channels, depth=5, **kwargs):
        super().__init__(**kwargs)
        self._depth = depth
        self._out_channels = out_channels
        self._in_channels = 3
        del self.classifier

    def get_stages(self):
        return [
            nn.Identity(),
            self.features[:2],
            self.features[2:4],
            self.features[4:7],
            self.features[7:14],
            self.features[14:],
        ]

    def forward(self, x):
        stages = self.get_stages()

        features = []
        for i in range(self._depth + 1):
            x = stages[i](x)
            features.append(x)

        return features

    def load_state_dict(self, state_dict, **kwargs):
        state_dict.pop("classifier.1.bias")
        state_dict.pop("classifier.1.weight")
        super().load_state_dict(state_dict, **kwargs)


mobilenet_encoders = {
    "mobilenet_v2": {
        "encoder": MobileNetV2Encoder,
        "pretrained_settings": {
            "imagenet": {
                "mean": [0.485, 0.456, 0.406],
                "std": [0.229, 0.224, 0.225],
                "url": "https://download.pytorch.org/models/mobilenet_v2-b0353104.pth",
                "input_space": "RGB",
                "input_range": [0, 1],
            },
        },
        "params": {
            "out_channels": (3, 16, 24, 32, 96, 1280),
        },
    },
}



import torch
import torch.nn as nn
import torch.nn.functional as F

from ..base import modules as md


class DecoderBlock(nn.Module):
    def __init__(
            self,
            in_channels,
            skip_channels,
            out_channels,
            use_batchnorm=True,
            attention_type=None,
    ):
        super().__init__()
        self.conv1 = md.Conv2dReLU(
            in_channels + skip_channels,
            out_channels,
            kernel_size=3,
            padding=1,
            use_batchnorm=use_batchnorm,
        )
        self.attention1 = md.Attention(attention_type, in_channels=in_channels + skip_channels)
        self.conv2 = md.Conv2dReLU(
            out_channels,
            out_channels,
            kernel_size=3,
            padding=1,
            use_batchnorm=use_batchnorm,
        )
        self.attention2 = md.Attention(attention_type, in_channels=out_channels)

    def forward(self, x, skip=None):
        x = F.interpolate(x, scale_factor=2, mode="nearest")
        if skip is not None:
            x = torch.cat([x, skip], dim=1)
            x = self.attention1(x)
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.attention2(x)
        return x


class CenterBlock(nn.Sequential):
    def __init__(self, in_channels, out_channels, use_batchnorm=True):
        conv1 = md.Conv2dReLU(
            in_channels,
            out_channels,
            kernel_size=3,
            padding=1,
            use_batchnorm=use_batchnorm,
        )
        conv2 = md.Conv2dReLU(
            out_channels,
            out_channels,
            kernel_size=3,
            padding=1,
            use_batchnorm=use_batchnorm,
        )
        super().__init__(conv1, conv2)


class UnetDecoder(nn.Module):
    def __init__(
            self,
            encoder_channels,
            decoder_channels,
            n_blocks=5,
            use_batchnorm=True,
            attention_type=None,
            center=False,
    ):
        super().__init__()

        if n_blocks != len(decoder_channels):
            raise ValueError(
                "Model depth is {}, but you provide `decoder_channels` for {} blocks.".format(
                    n_blocks, len(decoder_channels)
                )
            )

        encoder_channels = encoder_channels[1:]  # remove first skip with same spatial resolution
        encoder_channels = encoder_channels[::-1]  # reverse channels to start from head of encoder

        # computing blocks input and output channels
        head_channels = encoder_channels[0]
        in_channels = [head_channels] + list(decoder_channels[:-1])
        skip_channels = list(encoder_channels[1:]) + [0]
        out_channels = decoder_channels

        if center:
            self.center = CenterBlock(
                head_channels, head_channels, use_batchnorm=use_batchnorm
            )
        else:
            self.center = nn.Identity()

        # combine decoder keyword arguments
        kwargs = dict(use_batchnorm=use_batchnorm, attention_type=attention_type)
        blocks = [
            DecoderBlock(in_ch, skip_ch, out_ch, **kwargs)
            for in_ch, skip_ch, out_ch in zip(in_channels, skip_channels, out_channels)
        ]
        self.blocks = nn.ModuleList(blocks)

    def forward(self, *features):

        features = features[1:]    # remove first skip with same spatial resolution
        features = features[::-1]  # reverse channels to start from head of encoder

        head = features[0]
        skips = features[1:]

        x = self.center(head)
        for i, decoder_block in enumerate(self.blocks):
            skip = skips[i] if i < len(skips) else None
            x = decoder_block(x, skip)

        return x



import torch.nn as nn
from .modules import Flatten, Activation


class SegmentationHead(nn.Sequential):

    def __init__(self, in_channels, out_channels, kernel_size=3, activation=None, upsampling=1):
        conv2d = nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size, padding=kernel_size // 2)
        upsampling = nn.UpsamplingBilinear2d(scale_factor=upsampling) if upsampling > 1 else nn.Identity()
        activation = Activation(activation)
        super().__init__(conv2d, upsampling, activation)






from typing import Optional, Union, List
from .decoder import UnetDecoder
from ..encoders import get_encoder
from ..base import SegmentationModel
from ..base import SegmentationHead


class Unet(SegmentationModel):
    """Unet_ is a fully convolution neural network for image semantic segmentation

    Args:
        encoder_name: name of classification model (without last dense layers) used as feature
            extractor to build segmentation model.
        encoder_depth (int): number of stages used in decoder, larger depth - more features are generated.
            e.g. for depth=3 encoder will generate list of features with following spatial shapes
            [(H,W), (H/2, W/2), (H/4, W/4), (H/8, W/8)], so in general the deepest feature tensor will have
            spatial resolution (H/(2^depth), W/(2^depth)]
        encoder_weights: one of ``None`` (random initialization), ``imagenet`` (pre-training on ImageNet).
        decoder_channels: list of numbers of ``Conv2D`` layer filters in decoder blocks
        decoder_use_batchnorm: if ``True``, ``BatchNormalisation`` layer between ``Conv2D`` and ``Activation`` layers
            is used. If 'inplace' InplaceABN will be used, allows to decrease memory consumption.
            One of [True, False, 'inplace']
        decoder_attention_type: attention module used in decoder of the model
            One of [``None``, ``scse``]
        in_channels: number of input channels for model, default is 3.
        classes: a number of classes for output (output shape - ``(batch, classes, h, w)``).
        activation: activation function to apply after final convolution;
            One of [``sigmoid``, ``softmax``, ``logsoftmax``, ``identity``, callable, None]
        aux_params: if specified model will have additional classification auxiliary output
            build on top of encoder, supported params:
                - classes (int): number of classes
                - pooling (str): one of 'max', 'avg'. Default is 'avg'.
                - dropout (float): dropout factor in [0, 1)
                - activation (str): activation function to apply "sigmoid"/"softmax" (could be None to return logits)

    Returns:
        ``torch.nn.Module``: **Unet**

    .. _Unet:
        https://arxiv.org/pdf/1505.04597

    """

    def __init__(
        self,
        encoder_name: str = "resnet34",
        encoder_depth: int = 5,
        encoder_weights: str = "imagenet",
        decoder_use_batchnorm: bool = True,
        decoder_channels: List[int] = (256, 128, 64, 32, 16),
        decoder_attention_type: Optional[str] = None,
        in_channels: int = 3,
        classes: int = 1,
        activation: Optional[Union[str, callable]] = None,
        aux_params: Optional[dict] = None,
    ):
        super().__init__()

        self.encoder = get_encoder(
            encoder_name,
            in_channels=in_channels,
            depth=encoder_depth,
            weights=encoder_weights,
        )

        self.mask_decoder = UnetDecoder(
            encoder_channels=self.encoder.out_channels,
            decoder_channels=decoder_channels,
            n_blocks=encoder_depth,
            use_batchnorm=decoder_use_batchnorm,
            center=True if encoder_name.startswith("vgg") else False,
            attention_type=decoder_attention_type,
        )

        self.mask_segmentation_head = SegmentationHead(
            in_channels=decoder_channels[-1],
            out_channels=classes,
            activation=activation,
            kernel_size=3,
        )

        self.contour_decoder = UnetDecoder(
            encoder_channels=self.encoder.out_channels,
            decoder_channels=decoder_channels,
            n_blocks=encoder_depth,
            use_batchnorm=decoder_use_batchnorm,
            center=True if encoder_name.startswith("vgg") else False,
            attention_type=decoder_attention_type,
        )

        self.contour_segmentation_head = SegmentationHead(
            in_channels=decoder_channels[-1],
            out_channels=classes,
            activation=activation,
            kernel_size=3,
        )

        

        self.name = "u-{}".format(encoder_name)
        self.initialize()



import re
import functools
import torch
import torch.nn as nn


class Activation(nn.Module):
    def __init__(self, activation):
        super().__init__()
        if activation == None or activation == 'identity':
            self.activation = nn.Identity()
        elif activation == 'sigmoid':
            self.activation = torch.sigmoid
        elif activation == 'softmax2d':
            self.activation = functools.partial(torch.softmax, dim=1)
        elif callable(activation):
            self.activation = activation
        else:
            raise ValueError

    def forward(self, x):
        return self.activation(x)

class BaseObject(nn.Module):

    def __init__(self, name=None):
        super().__init__()
        self._name = name

    @property
    def __name__(self):
        if self._name is None:
            name = self.__class__.__name__
            s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
            return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()
        else:
            return self._name


class Metric(BaseObject):
    pass


class Loss(BaseObject):

    def __add__(self, other):
        if isinstance(other, Loss):
            return SumOfLosses(self, other)
        else:
            raise ValueError('Loss should be inherited from `Loss` class')

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, value):
        if isinstance(value, (int, float)):
            return MultipliedLoss(self, value)
        else:
            raise ValueError('Loss should be inherited from `BaseLoss` class')

    def __rmul__(self, other):
        return self.__mul__(other)


class SumOfLosses(Loss):

    def __init__(self, l1, l2):
        name = '{} + {}'.format(l1.__name__, l2.__name__)
        super().__init__(name=name)
        self.l1 = l1
        self.l2 = l2

    def __call__(self, *inputs):
        return self.l1.forward(*inputs) + self.l2.forward(*inputs)


class MultipliedLoss(Loss):

    def __init__(self, loss, multiplier):

        # resolve name
        if len(loss.__name__.split('+')) > 1:
            name = '{} * ({})'.format(multiplier, loss.__name__)
        else:
            name = '{} * {}'.format(multiplier, loss.__name__)
        super().__init__(name=name)
        self.loss = loss
        self.multiplier = multiplier

    def __call__(self, *inputs):
        return self.multiplier * self.loss.forward(*inputs)



import torch.nn as nn
import torch
from . import base
from . import functional as F
from  .base import Activation


class JaccardLoss(base.Loss):

    def __init__(self, eps=1., activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        y_pr = self.activation(y_pr)
        return 1 - F.jaccard(
            y_pr, y_gt,
            eps=self.eps,
            threshold=None,
            ignore_channels=self.ignore_channels,
        )


class DiceLoss(base.Loss):

    def __init__(self, eps=1., beta=1., activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps
        self.beta = beta
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        y_pr = self.activation(y_pr)
        return 1 - F.f_score(
            y_pr, y_gt,
            beta=self.beta,
            eps=self.eps,
            threshold=None,
            ignore_channels=self.ignore_channels,
        )

class TverskyLoss(base.Loss):

    def __init__(self, alpha=0.3,  eps=1., beta=0.7, activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps
        self.alpha = alpha
        self.beta = beta
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        y_pr = self.activation(y_pr)
        return 1 - F.tversky_score(
            y_pr, y_gt,
            alpha=self.alpha,
            beta=self.beta,
            eps=self.eps,
            threshold=None,
            ignore_channels=self.ignore_channels,
        )

class FocalTverskyLoss(base.Loss):

    def __init__(self, alpha=0.3,  eps=1., beta=0.7, gamma=0.75,  activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        y_pr = self.activation(y_pr)
        tversky_loss =  1 - F.tversky_score(
            y_pr, y_gt,
            alpha=self.alpha,
            beta=self.beta,
            eps=self.eps,
            threshold=None,
            ignore_channels=self.ignore_channels,
        )

        return torch.pow(tversky_loss, self.gamma)


class L1Loss(nn.L1Loss, base.Loss):
    pass


class MSELoss(nn.MSELoss, base.Loss):
    pass


class CrossEntropyLoss(nn.CrossEntropyLoss, base.Loss):
    pass


class NLLLoss(nn.NLLLoss, base.Loss):
    pass


class BCELoss(nn.BCELoss, base.Loss):
    pass


class BCEWithLogitsLoss(nn.BCEWithLogitsLoss, base.Loss):
    def __init__(self, activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)

        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        y_pr = self.activation(y_pr)
        return F.binary_cross_entropy_with_logits(
            y_pr, y_gt,
            threshold=None,
            ignore_channels=self.ignore_channels,
        )



import torch
import torch.nn as nn

try:
    from inplace_abn import InPlaceABN
except ImportError:
    InPlaceABN = None


class Conv2dReLU(nn.Sequential):
    def __init__(
            self,
            in_channels,
            out_channels,
            kernel_size,
            padding=0,
            stride=1,
            use_batchnorm=True,
    ):

        if use_batchnorm == "inplace" and InPlaceABN is None:
            raise RuntimeError(
                "In order to use `use_batchnorm='inplace'` inplace_abn package must be installed. "
                + "To install see: https://github.com/mapillary/inplace_abn"
            )

        conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            bias=not (use_batchnorm),
        )
        relu = nn.ReLU(inplace=True)

        if use_batchnorm == "inplace":
            bn = InPlaceABN(out_channels, activation="leaky_relu", activation_param=0.0)
            relu = nn.Identity()

        elif use_batchnorm and use_batchnorm != "inplace":
            bn = nn.BatchNorm2d(out_channels)

        else:
            bn = nn.Identity()

        super(Conv2dReLU, self).__init__(conv, bn, relu)


class SCSEModule(nn.Module):
    def __init__(self, in_channels, reduction=16):
        super().__init__()
        self.cSE = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(in_channels, in_channels // reduction, 1),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels // reduction, in_channels, 1),
            nn.Sigmoid(),
        )
        self.sSE = nn.Sequential(nn.Conv2d(in_channels, 1, 1), nn.Sigmoid())

    def forward(self, x):
        return x * self.cSE(x) + x * self.sSE(x)


class ArgMax(nn.Module):

    def __init__(self, dim=None):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        return torch.argmax(x, dim=self.dim)


class Activation(nn.Module):

    def __init__(self, name, **params):

        super().__init__()

        if name is None or name == 'identity':
            self.activation = nn.Identity(**params)
        elif name == 'sigmoid':
            self.activation = nn.Sigmoid()
        elif name == 'softmax2d':
            self.activation = nn.Softmax(dim=1, **params)
        elif name == 'softmax':
            self.activation = nn.Softmax(**params)
        elif name == 'logsoftmax':
            self.activation = nn.LogSoftmax(**params)
        elif name == 'argmax':
            self.activation = ArgMax(**params)
        elif name == 'argmax2d':
            self.activation = ArgMax(dim=1, **params)
        elif callable(name):
            self.activation = name(**params)
        else:
            raise ValueError('Activation should be callable/sigmoid/softmax/logsoftmax/None; got {}'.format(name))

    def forward(self, x):
        return self.activation(x)


class Attention(nn.Module):

    def __init__(self, name, **params):
        super().__init__()

        if name is None:
            self.attention = nn.Identity(**params)
        elif name == 'scse':
            self.attention = SCSEModule(**params)
        else:
            raise ValueError("Attention {} is not implemented".format(name))

    def forward(self, x):
        return self.attention(x)


class Flatten(nn.Module):
    def forward(self, x):
        return x.view(x.shape[0], -1)
